'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.WorkerDiscovery = undefined;

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _stream = require('../utils/stream');

var _outside = require('./worker/outside');

var _bitcoinjsLibZcash = require('bitcoinjs-lib-zcash');

var _simpleWorkerChannel = require('../utils/simple-worker-channel');

var _addressSource = require('../address-source');

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var WorkerDiscovery = exports.WorkerDiscovery = function () {
    function WorkerDiscovery(discoveryWorkerFactory, fastXpubWorker, fastXpubWasmPromise, chain) {
        var _this = this;

        _classCallCheck(this, WorkerDiscovery);

        this.forceAddedTransactions = [];
        this.forceAddedTransactionsEmitter = new _stream.Emitter();
        this.forceAddedTransactionsStream = _stream.Stream.fromEmitter(this.forceAddedTransactionsEmitter, function () {}).map(function () {
            return 'block';
        });

        this.discoveryWorkerFactory = discoveryWorkerFactory;
        // $FlowIssue
        this.addressWorkerChannel = typeof WebAssembly === 'undefined' ? null : new _simpleWorkerChannel.WorkerChannel(fastXpubWorker);
        fastXpubWasmPromise.then(function (binary) {
            if (_this.addressWorkerChannel !== null) {
                fastXpubWorker.postMessage({ type: 'init', binary: binary });
            }
        }, function (error) {
            return console.error(error);
        });
        this.chain = chain;
    }

    _createClass(WorkerDiscovery, [{
        key: 'tryHDNode',
        value: function tryHDNode(xpub, network) {
            try {
                var node = _bitcoinjsLibZcash.HDNode.fromBase58(xpub, network, true);
                return node;
            } catch (e) {
                return e;
            }
        }
    }, {
        key: 'forceAddTransaction',


        // useful for adding transactions right after succesful send
        value: function forceAddTransaction(transaction) {
            this.forceAddedTransactions.push(transaction);
            this.forceAddedTransactionsEmitter.emit(true);
        }
    }, {
        key: 'detectUsedAccount',
        value: function detectUsedAccount(xpub, network, segwit) {
            var _this2 = this;

            var node = this.tryHDNode(xpub, network);
            if (node instanceof Error) {
                return Promise.reject(node);
            }
            var external = node.derive(0);
            var internal = node.derive(1);

            var sources = [this.createAddressSource(external, network, segwit), this.createAddressSource(internal, network, segwit)];
            // I do not actually need to do any logic in the separate worker for discovery

            var GAP_SIZE = 20;
            return Promise.all([_outside.WorkerDiscoveryHandler.deriveAddresses(sources[0], null, 0, GAP_SIZE - 1), _outside.WorkerDiscoveryHandler.deriveAddresses(sources[1], null, 0, GAP_SIZE - 1)]).then(function (_ref) {
                var _ref2 = _slicedToArray(_ref, 2),
                    addressesA = _ref2[0],
                    addressesB = _ref2[1];

                return _this2.chain.lookupTransactionsIds(addressesA.concat(addressesB), 100000000, 0);
            }).then(function (ids) {
                return ids.length !== 0;
            });
        }
    }, {
        key: 'discoverAccount',
        value: function discoverAccount(initial, xpub, network, segwit) {
            var node = this.tryHDNode(xpub, network);
            if (node instanceof Error) {
                return _stream.StreamWithEnding.fromStreamAndPromise(_stream.Stream.fromArray([]), Promise.reject(node));
            }
            var external = node.derive(0);
            var internal = node.derive(1);

            var sources = [this.createWorkerAddressSource(external, network, segwit), this.createWorkerAddressSource(internal, network, segwit)];

            var out = new _outside.WorkerDiscoveryHandler(this.discoveryWorkerFactory, this.chain, sources, network, this.forceAddedTransactions);
            return out.discovery(initial, xpub, segwit === 'p2sh');
        }
    }, {
        key: 'monitorAccountActivity',
        value: function monitorAccountActivity(initial, xpub, network, segwit) {
            var _this3 = this;

            var node = this.tryHDNode(xpub, network);
            if (node instanceof Error) {
                return _stream.Stream.simple(node);
            }
            var external = node.derive(0);
            var internal = node.derive(1);

            var sources = [this.createWorkerAddressSource(external, network, segwit), this.createWorkerAddressSource(internal, network, segwit)];

            function allAddresses(info) {
                return new Set(info.usedAddresses.map(function (a) {
                    return a.address;
                }).concat(info.unusedAddresses).concat(info.changeAddresses));
            }

            this.chain.subscribe(allAddresses(initial));
            var currentState = initial;

            var txNotifs = this.chain.notifications.filter(function (tx) {
                // determine if it's mine
                var addresses = allAddresses(currentState);
                var mine = false;
                tx.inputAddresses.concat(tx.outputAddresses).forEach(function (a) {
                    if (a != null) {
                        if (addresses.has(a)) {
                            mine = true;
                        }
                    }
                });
                return mine;
            })
            // flow thing
            .map(function (tx) {
                return tx;
            });

            // we need to do updates on blocks, if there are unconfs
            var blockStream = this.chain.blocks.map(function () {
                return 'block';
            });

            var resNull = blockStream.concat(txNotifs).concat(this.forceAddedTransactionsStream).mapPromiseError(function () {
                var out = new _outside.WorkerDiscoveryHandler(_this3.discoveryWorkerFactory, _this3.chain, sources, network, _this3.forceAddedTransactions);
                return out.discovery(currentState, xpub, segwit === 'p2sh').ending.then(function (res) {
                    currentState = res;
                    return res;
                });
            });

            var res = _stream.Stream.filterNull(resNull);
            return res;
        }
    }, {
        key: 'createAddressSource',
        value: function createAddressSource(node, network, segwit) {
            var source = this.createWorkerAddressSource(node, network, segwit);
            if (source == null) {
                return new _addressSource.BrowserAddressSource(node, network, segwit === 'p2sh');
            }
            return source;
        }
    }, {
        key: 'createWorkerAddressSource',
        value: function createWorkerAddressSource(node, network, segwit) {
            var addressWorkerChannel = this.addressWorkerChannel;
            if (addressWorkerChannel == null) {
                return null;
            }
            var version = segwit === 'p2sh' ? network.scriptHash : network.pubKeyHash;
            return new _addressSource.WorkerAddressSource(addressWorkerChannel, node, version, segwit);
        }
    }]);

    return WorkerDiscovery;
}();